"""
Various web page clicks

TODO:
 - This will grow rapidly, so look into ways of splitting up
 - The hardcoded URLs are fragile
"""

from django.contrib.auth.models import User
from django.test import TestCase

from wikiwikiweb.models import WikiPage


class TestBasicCalls(TestCase):
    """General control flow through site"""

    # TODO: can we move the files nearer?
    fixtures = ['basic-db.json']

    #
    # Potential TODO:
    #  - force Space for create page - done but lacks test
    #  - check rev matches WikiName for archive stuff (not urgent, just seems odd right now) See #11
    #  -
    #

    def setUp(self):
        self._test_user = 'testuser'
        self._test_pass = 'testpass123'
        self._sess_key_space = 'sess_space_key'

    # Home page ---------------------------------------------------------------

    def test_homepage_load(self):
        response = self.client.get('/')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wiki/home.html')

    # WikiPage pages-----------------------------------------------------------

    # View page

    def test_success_for_real_wikipage(self):
        response = self.client.get('/AvrilLavigne')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wiki/wiki_page.html')

    # Create page

    def test_require_login_for_create_not_logged_in(self):
        response = self.client.get('/ThisNoExist')
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response,
                        '/accounts/login/?next=/ThisNoExist', # Seems fragile
                        fetch_redirect_response=False)

    def test_get_create_form_if_logged_in(self):
        self._populate_session()

        response = self.client.get('/ImaginaryPage')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wiki/wiki_page_create.html')

    # TODO: test invalid form data

    def test_create_form_post(self):
        self._populate_session()

        newpagename = 'MyNewPage' # nb. not in fixture db

        # nb. 'name' will soon be retired, see #18
        formdata = {'name': 'This name is a red herring',
                    'content': 'Some page content',
                    'edit_reason': 'Created1'}

        newpageurl = '/' + newpagename

        response = self.client.post(newpageurl, data=formdata)
        self.assertRedirects(response,
                        newpageurl + '?success=created',
                        fetch_redirect_response=False)

        newpage = WikiPage.objects.get(pk=newpagename)

        self.assertEqual(newpage.created_by, self._get_user(self._test_user))
        self.assertEqual(newpage.updated_by, self._get_user(self._test_user))

    # Edit page

    def test_get_edit_form_if_logged_in(self):
        self._populate_session()

        response = self.client.get('/AvrilLavigne/edit')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wiki/wiki_page_edit.html')

    def test_edit_form_post(self):
        self._populate_session()

        editpagename = 'JoniMitchell' # nb. is in fixture db; created by user 4/otheruser

        formdata = {
                    'content': 'Some edited page content',
                    'edit_reason': 'Edited1'
                    }

        viewpageurl = '/' + editpagename
        editpageurl = viewpageurl + '/edit'

        response = self.client.post(editpageurl, data=formdata)
        self.assertRedirects(response,
                            viewpageurl + '?success=updated',
                            fetch_redirect_response=False)

        newpage = WikiPage.objects.get(pk=editpagename)

        # check this hasn't changed...
        self.assertEqual(newpage.created_by, self._get_user('otheruser'))

        # ...but this has
        self.assertEqual(newpage.updated_by, self._get_user(self._test_user))

    # Archive pages -----------------------------------------------------------

    def test_success_for_real_archivepage(self):
        """Checks for #22"""

        response = self.client.get('/JacksonBrowne/86')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wiki/wiki_page_archive.html')

    def test_for_mismatched_revision_id(self):
        """Checks for #11"""

        # 86 would be JacksonBrowne, not AvriLavigne
        response = self.client.get('/AvriLavigne/86')
        self.assertEqual(response.status_code, 404)

    # WikiSpace pages ---------------------------------------------------------

    def test_success_for_real_wikispace(self):
        response = self.client.get('/space:ThreeHundredSongs')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wiki/space.html')

    def test_selectspace_form_post(self):

        session = self.client.session
        self.assertFalse(session.get(self._sess_key_space))

        targetpage = '/search?q=Kylie'
        selectpage = '/space/select'

        formdata = {
                    'space_choice': 1,
                    'next': targetpage
                }

        response = self.client.post(selectpage, data=formdata)
        self.assertRedirects(response,
                        targetpage,
                        fetch_redirect_response=False)

    # Search page -------------------------------------------------------------

    def test_request_search_with_space_in_session(self):

        self._populate_session()

        response = self.client.get('/search?q=avril')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wiki/search.html')

    def test_request_search_with_no_space(self):

        session = self.client.session
        self.assertFalse(session.get(self._sess_key_space))

        response = self.client.get('/search?q=Avril+Lavigne')
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, '/space/select?next=/search?q=Avril+Lavigne')

    # User page ---------------------------------------------------------------

    def test_success_for_real_userpage(self):
        response = self.client.get('/user:testuser')
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, 'wiki/user.html')

    # Private methods ----------------------------------------------------------

    def _populate_session(self):
        """Ensure user logged in, and valid WikiSpace in session"""

        self.client.login(username=self._test_user, password=self._test_pass)

        session = self.client.session
        session[self._sess_key_space] = 'ThreeHundredSongs'
        session.save()

    def _get_user(self, username):
        return User.objects.get(username=username)



from django.contrib import admin

from .models import *


@admin.register(WikiPage)
class WikiPageAdmin(admin.ModelAdmin):

    list_display = ['name', 'space', 'updated']

    fields = [
            'name',
            'space',
            'content',
            'edit_reason',
            'created_by',
            'updated_by',
    ]

    def get_readonly_fields(self, request, obj=None):
        if obj: # editing an existing object
            return self.readonly_fields + ('name',)
        return self.readonly_fields


@admin.register(WikiSpace)
class WikiSpaceAdmin(admin.ModelAdmin):

    list_display = ['name', 'homepage']

    fields = [
            'name',
            'description',
            'homepage',
    ]

    def get_form(self, request, obj=None, **kwargs):
        """Customise the homepage widget"""

        form = super(WikiSpaceAdmin, self).get_form(request, obj, **kwargs)
        field = form.base_fields['homepage']

        # Disable spurious/troublesome actions
        field.widget.can_add_related = False
        field.widget.can_view_related = False
        field.widget.can_change_related = False
        field.widget.can_delete_related = False

        # Filter options to this space
        field.queryset = WikiPage.objects.filter(space=obj)

        return form



# Generated by Django 5.0.6 on 2024-06-28 15:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wikiwikiweb', '0002_historicalwikispace_homepage_wikispace_homepage'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalwikipage',
            name='created_by',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='historicalwikipage',
            name='updated_by',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='wikipage',
            name='created_by',
            field=models.ForeignKey(default=1, on_delete=django.db.models.deletion.PROTECT, related_name='creator', to=settings.AUTH_USER_MODEL),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='wikipage',
            name='updated_by',
            field=models.ForeignKey(default=1, on_delete=django.db.models.deletion.PROTECT, related_name='updater', to=settings.AUTH_USER_MODEL),
            preserve_default=False,
        ),
    ]



from django.contrib.auth.models import User
from django.db import models
from simple_history.models import HistoricalRecords


class WikiSpace(models.Model):

    spaceid = models.AutoField(primary_key=True)
    name = models.SlugField(default='', null=False)
    description = models.TextField(blank=True, null=True)
    homepage = models.ForeignKey('WikiPage', blank=True, null=True, on_delete=models.PROTECT)

    history = HistoricalRecords()

    class Meta:
        verbose_name = 'Wiki Space'
        ordering = ['name']

    def __str__(self):
        return self.name


class WikiPage(models.Model):

    name = models.CharField('Page Name', max_length=255, primary_key=True)
    space = models.ForeignKey(WikiSpace, on_delete=models.PROTECT, default=1)
    content = models.TextField('Page Content', blank=False, null=False)
    created = models.DateTimeField(blank=False, null=False, auto_now_add=True)
    updated = models.DateTimeField(blank=False, null=False, auto_now=True)
    created_by = models.ForeignKey(User, on_delete=models.PROTECT, related_name='pagescreated')
    updated_by = models.ForeignKey(User, on_delete=models.PROTECT, related_name='pagesupdated')

    #updated = models.DateTimeField(blank=False, null=False, auto_now=True)
    edit_reason = models.CharField('Reason for Edit', max_length=256, default='Created')
    history = HistoricalRecords()

    class Meta:
        verbose_name = 'Wiki Page'

    def __str__(self):
        return self.name



import re

from django import template
from django.urls import reverse

from ..models import WikiPage


register = template.Library()

# I've tried to enforce https://wiki.c2.com/?WikiCase as far as sensible
# There will be omissions/exceptions
FMT_WIKINAME = r'(?<!=)\b[A-Z]+[a-zA-Z0-9]+[A-Z][a-zA-Z0-9]+\b'


@register.filter
def linkify(pagetext):

    wikinames = set(re.findall(FMT_WIKINAME, pagetext))

    for item in wikinames:
        my_regex = r"\b(?=\w)" + re.escape(item) + r"\b(?!\w)"
        pagetext = re.sub(my_regex, _make_link(item), pagetext)

    return pagetext


def _make_link(wikiname):

    try:
        obj = WikiPage.objects.get(name=wikiname)
        page_exists = True
    except:
        page_exists = False

    # TODO: proper string formatting

    url = reverse('djwiki:page_view', kwargs={'pk': wikiname})

    link = '<span class="wikilink"><a href="' + url + '"'

    if not page_exists:
        link = link + ' class="newpage"'

    link = link + '>' + wikiname + '</a></span>'

    return link

@register.filter
def make_title(wikiname):
    # See: https://stackoverflow.com/questions/199059/a-pythonic-way-to-insert-a-space-before-capital-letters
    # return re.sub(r"(\w)([A-Z])", r"\1 \2", wikiname)
    return re.sub(r"(?<=\w)([A-Z])", r" \1", wikiname)



# Generated by Django 5.0.6 on 2024-06-25 12:48

import django.db.models.deletion
import simple_history.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='WikiSpace',
            fields=[
                ('spaceid', models.AutoField(primary_key=True, serialize=False)),
                ('name', models.SlugField(default='')),
                ('description', models.TextField(blank=True, null=True)),
            ],
            options={
                'verbose_name': 'Wiki Space',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='HistoricalWikiSpace',
            fields=[
                ('spaceid', models.IntegerField(blank=True, db_index=True)),
                ('name', models.SlugField(default='')),
                ('description', models.TextField(blank=True, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField(db_index=True)),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical Wiki Space',
                'verbose_name_plural': 'historical Wiki Spaces',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': ('history_date', 'history_id'),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='WikiPage',
            fields=[
                ('name', models.CharField(max_length=255, primary_key=True, serialize=False, verbose_name='Page Name')),
                ('content', models.TextField(verbose_name='Page Content')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('edit_reason', models.CharField(default='Created', max_length=256, verbose_name='Reason for Edit')),
                ('space', models.ForeignKey(default=1, on_delete=django.db.models.deletion.PROTECT, to='wikiwikiweb.wikispace')),
            ],
            options={
                'verbose_name': 'Wiki Page',
            },
        ),
        migrations.CreateModel(
            name='HistoricalWikiPage',
            fields=[
                ('name', models.CharField(db_index=True, max_length=255, verbose_name='Page Name')),
                ('content', models.TextField(verbose_name='Page Content')),
                ('created', models.DateTimeField(blank=True, editable=False)),
                ('updated', models.DateTimeField(blank=True, editable=False)),
                ('edit_reason', models.CharField(default='Created', max_length=256, verbose_name='Reason for Edit')),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField(db_index=True)),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
                ('space', models.ForeignKey(blank=True, db_constraint=False, default=1, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='wikiwikiweb.wikispace')),
            ],
            options={
                'verbose_name': 'historical Wiki Page',
                'verbose_name_plural': 'historical Wiki Pages',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': ('history_date', 'history_id'),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
    ]



from django.forms import (
    Form, ModelChoiceField, ModelForm
)

from .models import WikiPage, WikiSpace


# TODO: consider whether there should even be a name field. Maybe the ONLY way
# people should get here is via viewing a non-existent (yet) page. Thus, the
# view wouldn't even need its own route in urls.py. Also we wouldn't need to
# think about name clashes
class PageCreateForm(ModelForm):

    class Meta:
        model = WikiPage
        fields = ['content', 'edit_reason']


class PageEditForm(ModelForm):

    def __init__(self, *args, **kwargs):
        """ There MUST be a simpler way to override a simple input value"""

        initial = kwargs.get('initial', {})
        initial['edit_reason'] = ''
        kwargs['initial'] = initial
        super().__init__(*args, **kwargs)

    class Meta:
        model = WikiPage
        fields = ['content', 'edit_reason']


class SpaceSelectForm(Form):

    space_choice = ModelChoiceField(
                        queryset=WikiSpace.objects.all(),
                        label='Please select a space:',
                    )



from django.urls import path

from . import views

app_name = 'djwiki'

urlpatterns = [
    path('', views.HomeView.as_view(), name='home'),
    path('<str:page_name>/<int:pk>', views.PageArchiveView.as_view(), name='page_archive'),
    path('space/select', views.SpaceSelectView.as_view(), name='space_select'),
    path('space:<slug:slug>', views.SpaceView.as_view(), name='space'),
    path('user:<str:profileusername>', views.UserProfileView.as_view(), name='user'),
    path('<str:pk>/edit', views.PageEditView.as_view(), name='page_edit'),
    path('search', views.SearchView.as_view(), name='search'),
    path('mywiki', views.MyView.as_view(), name='mywiki'), # unlinked so far

    # Always last, as we fall back to this by default
    path('<str:pk>', views.PageView.as_view(), name='page_view'),
]



SECRET_KEY = 'fake-key'

INSTALLED_APPS = [
    'tests',
    'wikiwikiweb',
    'markdownify',
    'simple_history',
    'django_bootstrap5',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
]

DATABASES={
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}

ROOT_URLCONF='tests.urls'

MIDDLEWARE = [
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'wikiwikiweb.context_processors.site_wide_context',
            ],
            'debug': True,
        },
    },
]



# Generated by Django 5.0.6 on 2024-06-26 11:12

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wikiwikiweb', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalwikispace',
            name='homepage',
            field=models.ForeignKey(blank=True, db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='wikiwikiweb.wikipage'),
        ),
        migrations.AddField(
            model_name='wikispace',
            name='homepage',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='wikiwikiweb.wikipage'),
        ),
    ]



import re
from unittest import TestCase

# from django.test import TestCase

from wikiwikiweb.templatetags import wiki_extras


class LinkifyTestCase(TestCase):


    def setUp(self):
        self.tomatch = wiki_extras.FMT_WIKINAME


    def test_regex(self):
        """Test the regex independently of the code"""

        self.assertFalse(re.search(self.tomatch, ''))
        self.assertFalse(re.search(self.tomatch, 'hello world'))
        self.assertFalse(re.search(self.tomatch, 'hello worldOnAString'))

        self.assertTrue(re.search(self.tomatch, 'WikiName'))
        self.assertTrue(re.search(self.tomatch, 'this page has a WikiName in it'))
        self.assertTrue(re.search(self.tomatch, 'this WikiPage has two WikiNames in it'))

        self.assertEqual(re.findall(self.tomatch, 'WikiName'), ['WikiName'])
        self.assertEqual(re.findall(self.tomatch, 'this WikiPage has two WikiNames in it'), ['WikiPage', 'WikiNames'])


    def test_cannot_start_with_numbers(self):
        self.assertFalse(re.search(self.tomatch, 'hello 333WorldOnAString'))

    # I'll leave this for now. I'm not convinced it's a problem
    #def test_each_capital_letter_must_be_followed_by_a_lower_case_letter(self):
    #    self.assertFalse(re.search(self.tomatch, 'WhyUsePHP'));

    def test_linkify(self):
        self.assertFalse(self._is_linked(wiki_extras.linkify(''), 'whatever'))
        self.assertTrue(self._is_linked(wiki_extras.linkify('WikiPage'), 'WikiPage'))
        self.assertTrue(self._is_linked(
                wiki_extras.linkify('This is some text linking to WikiOtherPage somehow'),
                'WikiOtherPage')
            )


    def test_youtube_urls_unmolested(self):
        """YouTube vid IDs starting with upper case were erroneously linked"""

        youtube_url = 'https://www.youtube.com/watch?v=DVPab4bTGqY'

        self.assertFalse(re.search(self.tomatch, youtube_url))
        self.assertEqual(wiki_extras.linkify(youtube_url), youtube_url)

        self.assertFalse(self._is_linked(
                wiki_extras.linkify('https://www.youtube.com/watch?v=DVPab4bTGqY'),
                'DVPab4bTGqY')
            )


    def _is_linked(self, html, wikiname):
        """Don't get too specific about the HTML - that could get fragile"""

        return ('href="' in html
                and 'wikilink' in html
                and 'href="'  in html
                and '>' + wikiname + '</a>' in html)



from django.contrib.auth.models import User
from django.contrib.auth.decorators import login_required
from django.db.models import Q
from django.http import Http404
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views import generic

from .forms import *
from .models import *


ITEMS_PER_PAGE = 10

# TODO: check it actually exists. See #25
def wikispace_required(view_func):

    def wrapper(request, *args, **kwargs):
        if not request.session.get('sess_space_key'):
            intermediate = (reverse('djwiki:space_select')
                            + '?next='
                            + request.get_full_path())
            return redirect(intermediate)
        return view_func(request, *args, **kwargs)

    return wrapper


class HomeView(generic.TemplateView):
    template_name = 'wiki/home.html'

    def get_context_data(self, **kwargs):
        context = super(HomeView, self).get_context_data(**kwargs)
        context['pages_created'] = WikiPage.objects.order_by('-created')[:ITEMS_PER_PAGE]
        context['pages_updated'] = WikiPage.objects.order_by('-updated')[:ITEMS_PER_PAGE]

        # Display success dialogue. eg. logged out
        status_key = self.request.GET.get('success')
        context['status_key'] = status_key

        return context

    def get(self, request, *args, **kwargs):
        # Clear selected space   (TODO: decide whether this is actually helpful)
        request.session['sess_space_key'] = None
        return super().get(request, *args, **kwargs)


class PageView(generic.DetailView):
    model = WikiPage
    context_object_name = 'mypage'
    template_name = 'wiki/wiki_page.html'

    def dispatch(self, request, *args, **kwargs ):
        try:
            # Don't fully understand why this is so wordy, but it seems to be needed
            self.object = super().get_object()
            context = self.get_context_data(object=self.object)

            # TODO: some sort of constant for these key strings
            request.session['sess_space_key'] = self.object.space.name

            # Display success dialogue if that's what just happened
            status_key = request.GET.get('success')
            context['status_key'] = status_key

            return self.render_to_response(context)
        except Http404:
            view = PageCreateView.as_view()
            return view(request, *args, **kwargs)


class PageArchiveView(generic.DetailView):

    # TODO: set Space into session?

    model = HistoricalWikiPage
    context_object_name = 'mypage'
    template_name = 'wiki/wiki_page_archive.html'

    def get_queryset(self):
        search_wikiname = self.kwargs['page_name']
        # Nb. the pk has already magically been applied in a DetailView
        return HistoricalWikiPage.objects.filter(name=search_wikiname)


@method_decorator(login_required, name='dispatch')
@method_decorator(wikispace_required, name='dispatch')
class PageCreateView(generic.edit.CreateView):
    model = WikiPage
    context_object_name = 'mypage'
    template_name = 'wiki/wiki_page_create.html'

    form_class = PageCreateForm

    def setup(self, request, *args, **kwargs):
        """Initialize attributes shared by all view methods"""

        self._new_page_name = kwargs['pk']
        space_key = request.session.get('sess_space_key')

        # TODO: Remove the try/catch:
        # - only here because the Space doesn't get found in >=1 test case (exceot it does!)
        # - should be totally redundant after #25
        try:
            self.space = WikiSpace.objects.get(name=space_key)
        except:
            pass

        return super(PageCreateView, self).setup(request, *args, **kwargs)

    # nb. should this actually be get_context_data?
    def get(self, request, *args, **kwargs):

        defaults = {
            # 'name': self.kwargs['pk'],
        }

        form = self.form_class(initial=defaults)
        context = {
            'form': form,
            'new_page_name': self._new_page_name
        }
        return render(request, self.template_name, context)

    def form_valid(self, form):
        """Add current space and user to new page"""

        form.instance.name = self._new_page_name
        form.instance.space = self.space
        form.instance.created_by = self.request.user
        form.instance.updated_by = self.request.user
        return super(PageCreateView, self).form_valid(form)


    def get_success_url(self):
        new_page_name = self.object.name
        return reverse('djwiki:page_view', kwargs={'pk': new_page_name}) + '?success=created'


@method_decorator(login_required, name='dispatch')
class PageEditView(generic.edit.UpdateView):
    model = WikiPage
    context_object_name = 'mypage'
    template_name = 'wiki/wiki_page_edit.html'
    form_class = PageEditForm

    def form_valid(self, form):
        """Add current space and user to new page"""

        form.instance.updated_by = self.request.user
        return super(PageEditView, self).form_valid(form)

    def get_success_url(self):
        new_page_name = self.object.name
        return reverse('djwiki:page_view', kwargs={'pk': new_page_name}) + '?success=updated'


@method_decorator(wikispace_required, name='dispatch')
class SearchView(generic.ListView):

    template_name = 'wiki/search.html'
    model = WikiPage
    context_object_name = 'pages_list'
    #paginate_by = ITEMS_PER_PAGE

    def get_queryset(self): #
        # TODO: needs checking for empty
        search_string = self.request.GET.get('q')
        search_space = self.request.session.get('sess_space_key')

        return WikiPage.objects.filter(
            Q(space__name=search_space) &
            (Q(name__icontains=search_string) | Q(content__icontains=search_string)),
        )

    def get_context_data(self, **kwargs):
        context = super(SearchView, self).get_context_data(**kwargs)

        # Add query string if present
        if self.request.GET.get('q'):
            # TODO: nb. context.update
            context['search_query'] = self.request.GET.get('q')

        return context


class SpaceView(generic.DetailView):

    model = WikiSpace
    slug_field = 'name'
    context_object_name = 'myspace'
    template_name = 'wiki/space.html'

    def get(self, request, *args, **kwargs):
        # Would already have 404d if string is not a valid space
        request.session['sess_space_key'] = kwargs['slug']
        return super().get(request, *args, **kwargs)


class SpaceSelectView(generic.edit.FormView):

    form_class = SpaceSelectForm
    template_name = 'wiki/space_select.html'

    def form_valid(self, form):

        chosen_id = self.request.POST.get('space_choice')
        chosen_space = WikiSpace.objects.get(pk=chosen_id)
        self.request.session['sess_space_key'] = chosen_space.name

        return super().form_valid(form)

    def get_success_url(self):
        return self.request.POST.get('next')

    def get_context_data(self, **kwargs):
        context = super(SpaceSelectView, self).get_context_data(**kwargs)
        context['next'] = self.request.GET.get('next')
        return context


@method_decorator(login_required, name='dispatch')
class MyView(generic.TemplateView):

    template_name = 'wiki/mywiki.html'


class UserProfileView(generic.DetailView):

    model = User
    context_object_name = 'myuser'
    template_name = 'wiki/user.html'

    def get_object(self):
        return User.objects.get(username=self.kwargs['profileusername'])

    def get_context_data(self, **kwargs):

        context = super(UserProfileView, self).get_context_data(**kwargs)
        context['pages_created'] = self.object.pagescreated.all().order_by('-created') #[:ITEMS_PER_PAGE]
        context['edits_made'] = WikiPage.history.filter(updated_by=self.object).order_by('-updated')

        #context['debugme'] = WikiPage.history.filter(updated_by=self.object).count
        #context['debugme'] = self.object

        return context



from .models import WikiSpace


def site_wide_context(request):

    return {
        'all_spaces': WikiSpace.objects.all().order_by('name'),
        'sess_space_key': request.session.get('sess_space_key'),
    }


from django.urls import include, path

# the app urls.py must be include()d to pick up the route namespace ie 'djwiki'

urlpatterns = [
    path('accounts/', include('django.contrib.auth.urls')),
    path('', include('wikiwikiweb.urls')),
]



from django.contrib.auth.models import User
from django.test import TestCase

from wikiwikiweb.models import WikiPage, WikiSpace


class TrivialModelTestCase(TestCase):

    fixtures = ['basic-db.json']

    def test_I_exist(self):
        self.assertEqual(1+2, 3)

    def test_to_str(self):

        user = User.objects.create_user(username='zebedee')

        wikiname = 'LedZeppelin'

        space = WikiSpace.objects.create(name=wikiname)
        self.assertEqual(space.__str__(), wikiname)

        page = WikiPage.objects.create(name=wikiname, space=space, created_by=user, updated_by=user)
        self.assertEqual(page.__str__(), wikiname)
